#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

__constant__ char constant_bias[65536];

template <typename scalar_t>
__global__ void mlp_forward_kernel_optimized(
    const scalar_t* __restrict__ input,
    const scalar_t* __restrict__ weight,
    scalar_t* __restrict__ output,
    const int batch_size,
    const int in_features,
    const int out_features) {

    const int row = blockIdx.x;
    const int col = blockIdx.y;
    const int tid = threadIdx.x;
    const int warp_id = tid / warpSize;
    const int lane_id = tid % warpSize;
    
    // Shared memory for partial sums
    extern __shared__ char shared_memory[];
    scalar_t* shared_data = reinterpret_cast<scalar_t*>(shared_memory);

    if (row < batch_size && col < out_features) {
        // Each thread processes multiple elements (4x unrolling)
        scalar_t sum = 0;
        const int stride = blockDim.x * 4;
        const int input_offset = row * in_features;
        const int weight_offset = col * in_features;

        // Main loop with 4x unrolling
        #pragma unroll 4
        for (int i = tid * 4; i < in_features; i += stride) {
            scalar_t temp_sum = 0;
            if (i + 3 < in_features) {
                temp_sum = input[input_offset + i] * weight[weight_offset + i]
                        + input[input_offset + i + 1] * weight[weight_offset + i + 1]
                        + input[input_offset + i + 2] * weight[weight_offset + i + 2]
                        + input[input_offset + i + 3] * weight[weight_offset + i + 3];
            } else {
                for (int j = 0; j < 4 && i + j < in_features; j++) {
                    temp_sum += input[input_offset + i + j] * weight[weight_offset + i + j];
                }
            }
            sum += temp_sum;
        }

        // Warp-level reduction using shuffle
        #pragma unroll
        for (int offset = warpSize/2; offset > 0; offset /= 2) {
            sum += __shfl_down_sync(0xffffffff, sum, offset);
        }

        // First thread in each warp writes to shared memory
        if (lane_id == 0) {
            shared_data[warp_id] = sum;
        }
        __syncthreads();

        // Final reduction across warps by first thread
        if (tid == 0) {
            scalar_t final_sum = 0;
            const int num_warps = (blockDim.x + warpSize - 1) / warpSize;
            
            #pragma unroll
            for (int i = 0; i < num_warps; i++) {
                final_sum += shared_data[i];
            }
            
            // Add bias and write result
            final_sum += reinterpret_cast<const scalar_t*>(constant_bias)[col];
            output[row * out_features + col] = final_sum;
        }
    }
}

template <typename scalar_t>
__global__ void relu_kernel(
    scalar_t* __restrict__ data,
    const int size) {
    
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    const int stride = blockDim.x * gridDim.x;
    
    for (int i = idx; i < size; i += stride) {
        data[i] = data[i] > 0 ? data[i] : 0;
    }
}

torch::Tensor mlp_cuda_forward(
    torch::Tensor input,
    std::vector<torch::Tensor> weights,
    std::vector<torch::Tensor> biases) {
    
    auto device = input.device();
    const int num_layers = weights.size();
    auto current = input;
    
    for (size_t i = 0; i < num_layers; i++) {
        const int batch_size = current.size(0);
        const int in_features = current.size(1);
        const int out_features = weights[i].size(0);
        
        auto output = torch::empty({batch_size, out_features}, 
                                 torch::dtype(current.dtype()).device(device));
        
        // Copy bias to constant memory
        AT_DISPATCH_FLOATING_TYPES(current.type(), "bias_copy", ([&] {
            size_t bias_bytes = out_features * sizeof(scalar_t);
            if (bias_bytes <= sizeof(constant_bias)) {
                cudaMemcpyToSymbol(constant_bias, biases[i].data_ptr<scalar_t>(), bias_bytes);
            }
        }));

        // Calculate optimal thread block size and shared memory
        const int thread_block_size = 256;  // Multiple of warpSize(32)
        const int shared_mem_size = (thread_block_size / 32) * sizeof(float);
        dim3 blocks(batch_size, out_features);
        
        AT_DISPATCH_FLOATING_TYPES(current.type(), "mlp_forward_kernel_optimized", ([&] {
            mlp_forward_kernel_optimized<scalar_t><<<blocks, thread_block_size, shared_mem_size>>>(
                current.data_ptr<scalar_t>(),
                weights[i].data_ptr<scalar_t>(),
                output.data_ptr<scalar_t>(),
                batch_size,
                in_features,
                out_features
            );
        }));
        
        if (i < num_layers - 1) {
            const int size = batch_size * out_features;
            const int thread_per_block = 256;
            const int num_blocks = (size + thread_per_block - 1) / thread_per_block;
            
            AT_DISPATCH_FLOATING_TYPES(output.type(), "relu_kernel", ([&] {
                relu_kernel<scalar_t><<<num_blocks, thread_per_block>>>(
                    output.data_ptr<scalar_t>(),
                    size
                );
            }));
        }
        
        current = output;
    }
    
    return current;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &mlp_cuda_forward, "MLP forward (CUDA)");
}